
#include <iostream>
#include <string>

#ifdef _SF2_WINDOWS
#include "foundation/platform/sdl/platform_sdl_windows.h"
#elif defined(_SF2_LINUX)
#include "foundation/platform/sdl/platform_sdl_linux.h"
#else
#include "foundation/platform/sdl/platform_sdl_macos.h"
#endif
#include "foundation/graphics/viewport.h"
#include "foundation/input/keyboard.h"
#include "foundation/input/mouse.h"
#include "runtime/editor/editor_facility.h"
#include "runtime/editor/keys/keyhook_setup.h"
#include "utils/event.h"
#include "utils/delegate.h"
#include "utils/utilities.h"
#include "utils/keyhookstore.h"

using namespace Foundation;
using namespace Editor;

void BuildResource()
{
	//Utility::MakeBinaryResourceIncludeFile("logo_test.png", "data_logo.h", "data_logo", "Resource");
}

int main(int argc, char* argv[])
{
	//BuildResource();
    
    // Create the platform
    #ifdef _SF2_WINDOWS
        PlatformSDLWindows platform(argc, argv);
    #elif defined(_SF2_LINUX)
        PlatformSDLLinux platform(argc, argv);
    #else
        PlatformSDLMacOS platform(argc, argv);
    #endif
    
	// Initialize SDL
	const int sdl_init_result = SDL_Init(SDL_INIT_TIMER | SDL_INIT_AUDIO | SDL_INIT_VIDEO);

	if (sdl_init_result < 0)
	{
		std::cout << "SDL initialization failed. SDL Error: " << SDL_GetError();

		SDL_Quit();
		return -1;
	}

	// Create viewport
	const int width = 1280;
	const int height = 720;

	Viewport viewport(width, height, std::string("SID Factory II"));

	// Make keyboard and mouse input instances
	Keyboard keyboard;
	Mouse mouse;

	// Key setup
	Editor::KeyHookSetup key_hook_setup;

	// Editor facility
	EditorFacility editor(&platform, &viewport, key_hook_setup.GetKeyHookStore());

	// Start editor
	editor.Start(argc > 1 ? argv[1] : nullptr);

	// Ticking test
	unsigned int last_tick = SDL_GetTicks();

	const unsigned int updates_per_second = 30;
	const unsigned int target_frame_time = 1000 / updates_per_second;

	// Listen for SDL events
	SDL_Event event;
	bool force_quit = false;

	SDL_EventState(SDL_DROPFILE, SDL_ENABLE);

	while (!editor.IsDone() && !force_quit)
	{
		// Get tick count and maintain delta time
		const unsigned int tick = SDL_GetTicks();
		const int delta_tick = tick - last_tick;

		// Collect keyboard and mouse events
		keyboard.BeginCollect();
		mouse.BeginCollect();

		while (SDL_PollEvent(&event))
		{
			switch (event.type)
			{
			case SDL_QUIT:
				editor.TryQuit();
				break;
			case SDL_DROPFILE:
				editor.TryLoad(std::string(event.drop.file));
				SDL_free(event.drop.file);
				break;
			case SDL_TEXTINPUT:
				keyboard.KeyText(event.text.text);
				break;
			case SDL_KEYDOWN:
				keyboard.KeyDown(event.key.keysym.sym);
				break;
			case SDL_KEYUP:
				keyboard.KeyUp(event.key.keysym.sym);
				break;
			case SDL_MOUSEWHEEL:
				mouse.PushMouseWheelChange(static_cast<int>(event.wheel.x), static_cast<int>(event.wheel.y));
				break;
			}
		}

		keyboard.EndCollect();
		mouse.EndCollect();

		// Update mouse
		mouse.Update(delta_tick);

		// Update editor
		editor.Update(keyboard, mouse, delta_tick);

		// Yield the process, if it is required
		const unsigned int ticks_passed = SDL_GetTicks() - tick;

		if (ticks_passed < target_frame_time)
			SDL_Delay(target_frame_time - ticks_passed);

		// Refresh last tick
		last_tick = tick;
	}

	// Stop editor
	editor.Stop();

	// Close down SDL
	SDL_Quit();

	return 0;
}
