#include "viewport.h"
#include "imanaged.h"
#include "textfield.h"
#include "drawfield.h"
#include "image.h"
#include "resources/data_char.h"
#include <assert.h>

namespace Foundation
{
	Viewport::Viewport(int inResolutionX, int inResolutionY, const std::string& inCaption)
		: m_ResolutionX(inResolutionX)
		, m_ResolutionY(inResolutionY)
		, m_Window(nullptr)
		, m_Renderer(nullptr)
		, m_Caption(inCaption)
	{
		m_Window = SDL_CreateWindow(inCaption.c_str(), SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, m_ResolutionX, m_ResolutionY, SDL_WINDOW_SHOWN);
		assert(m_Window != nullptr);

		m_Renderer = SDL_CreateRenderer(m_Window, -1, 0);
		assert(m_Renderer != nullptr);
	}


	Viewport::~Viewport()
	{
		for (IManaged* text_field : m_ManagedResources)
			delete text_field;

		SDL_DestroyRenderer(m_Renderer);
		SDL_DestroyWindow(m_Window);
	}


	int Viewport::GetWidth() const
	{
		return m_ResolutionX;
	}


	int Viewport::GetHeight() const
	{
		return m_ResolutionY;
	}


	void Viewport::SetAdditionTitleInfo(const std::string& inAdditionTitleInfo)
	{
		if (inAdditionTitleInfo.empty())
			SDL_SetWindowTitle(m_Window, m_Caption.c_str());
		else
		{
			std::string title_with_addition_info = inAdditionTitleInfo + " - " + m_Caption;
			SDL_SetWindowTitle(m_Window, title_with_addition_info.c_str());
		}
	}



	void Viewport::Begin()
	{
		SDL_SetRenderDrawColor(m_Renderer, 0, 0, 0, 255);
		SDL_RenderClear(m_Renderer);

		for (auto text_field : m_ManagedResources)
			text_field->Begin();
	}


	void Viewport::End()
	{
		for (auto text_field : m_ManagedResources)
			text_field->End();

		SDL_RenderPresent(m_Renderer);
	}


	const Palette& Viewport::GetPalette() const
	{
		return m_Palette;
	}


	SDL_Window* Viewport::GetSDLWindow() const
	{
		return m_Window;
	}


	TextField* Viewport::CreateTextField(unsigned inWidth, unsigned int inHeight, int inX, int inY)
	{
		TextField* text_field = new TextField(*this, m_Renderer, inWidth, inHeight, inX, inY);

		m_ManagedResources.push_back(text_field);
		return text_field;
	}


	DrawField* Viewport::CreateDrawField(unsigned inWidth, unsigned int inHeight, int inX, int inY)
	{
		DrawField* draw_field = new DrawField(*this, m_Renderer, inWidth, inHeight, inX, inY);

		m_ManagedResources.push_back(draw_field);
		return draw_field;
	}


	Image* Viewport::CreateImageFromFile(const std::string& inFileName)
	{
		SDL_Surface* surface = SDL_LoadBMP(inFileName.c_str());

		if (surface != nullptr)
		{
			Image* image = new Image(*this, m_Renderer, surface);

			m_ManagedResources.push_back(image);
			return image;
		}

		return nullptr;
	}


	Image* Viewport::CreateImageFromARGBData(void* inData, unsigned int inWidth, unsigned int inHeight, bool inIncludeAlphaChannel)
	{
		unsigned int mask_r = 0x000000ff;
		unsigned int mask_g = 0x0000ff00;
		unsigned int mask_b = 0x00ff0000;
		unsigned int mask_a = 0xff000000;

		int depth = inIncludeAlphaChannel ? 32 : 24;
		int pitch = inWidth * (inIncludeAlphaChannel ? 4 : 3);

		SDL_Surface* surface = SDL_CreateRGBSurfaceFrom(inData, inWidth, inHeight, depth, pitch, mask_r, mask_g, mask_b, mask_a);
		Image* image = new Image(*this, m_Renderer, surface);

		m_ManagedResources.push_back(image);
		return image;
	}



	void Viewport::Destroy(IManaged* inManaged)
	{
		auto it = m_ManagedResources.begin();

		while (it != m_ManagedResources.end())
		{
			if (*it == inManaged)
			{
				m_ManagedResources.erase(it);
				delete inManaged;

				return;
			}

			++it;
		}
	}
}