#include "component_button.h"

#include "foundation/graphics/textfield.h"
#include "foundation/input/mouse.h"
#include "foundation/input/keyboard.h"
#include "foundation/input/keyboard_utils.h"
#include "runtime/editor/cursor_control.h"

#include "SDL_keycode.h"
#include <assert.h>

using namespace Foundation;

namespace Editor
{
	ComponentButton::ComponentButton(int inID, int inGroupID, Undo* inUndo, Foundation::TextField* inTextField, const std::string& inButtonText, int inX, int inY, int inWidth, std::function<void(void)> inButtonPressedCallback)
		: ComponentBase(inID, inGroupID, inUndo, inTextField, inX, inY, inWidth, 1)
		, m_ButtonText(inButtonText)
		, m_ButtonPressedCallback(inButtonPressedCallback)
		, m_MouseOver(false)
	{
		assert(inTextField != nullptr);
		assert(inButtonText.length() <= static_cast<size_t>(inWidth));
	}

	ComponentButton::~ComponentButton()
	{

	}

	//----------------------------------------------------------------------------------------------------------------------------------------

	void ComponentButton::SetHasControl(GetControlType inGetControlType, CursorControl& inCursorControl)
	{
		m_HasControl = true;
		inCursorControl.SetEnabled(false);
		m_RequireRefresh = true;
	}


	bool ComponentButton::ConsumeInput(const Foundation::Keyboard& inKeyboard, CursorControl& inCursorControl, ComponentsManager& inComponentsManager)
	{
		if (m_HasControl)
		{
			// Get key events
			for (const auto& key_event : inKeyboard.GetKeyEventList())
			{
				switch (key_event)
				{
				case SDLK_RETURN:
					m_ButtonPressedCallback();
					return true;
				default:
					break;
				}
			}
		}

		return false;
	}


	bool ComponentButton::ConsumeInput(const Foundation::Mouse& inMouse, bool inModifierKeyMask, CursorControl& inCursorControl, ComponentsManager& inComponentsManager)
	{
		if(m_MouseOver)
		{
			if (inMouse.IsButtonPressed(Foundation::Mouse::Button::Left))
			{
				m_ButtonPressedCallback();
				return true;
			}
		}

		return false;
	}


	void ComponentButton::ConsumeNonExclusiveInput(const Foundation::Mouse& inMouse)
	{
		Foundation::Point cell_position = m_TextField->GetCellPositionFromPixelPosition(inMouse.GetPosition());

		const bool mouse_over = (cell_position.m_Y == m_Position.m_Y && cell_position.m_X >= m_Position.m_X && cell_position.m_X < m_Position.m_X + m_Dimensions.m_Width);

		if (m_MouseOver != mouse_over)
			m_RequireRefresh = true;

		m_MouseOver = mouse_over;
	}

	//----------------------------------------------------------------------------------------------------------------------------------------

	void ComponentButton::Refresh(const DisplayState& inDisplayState)
	{
		if (m_RequireRefresh && m_TextField->IsEnabled())
		{
			Color color = [&]()
			{
				if (m_MouseOver)
					return m_HasControl ? Color::LightRed : Color::DarkGrey;
				else
					return m_HasControl ? Color::Red : Color::Black;
			}();
			m_TextField->ColorAreaBackground(color, m_Rect);

			int text_x = (m_Dimensions.m_Width - static_cast<int>(m_ButtonText.length())) >> 1;
			m_TextField->Print(m_Position.m_X + text_x, m_Position.m_Y, m_ButtonText);
		}
	}


	void ComponentButton::HandleDataChange()
	{
		if (m_HasDataChange)
			m_HasDataChange = false;
	}


	void ComponentButton::PullDataFromSource()
	{

	}



	void ComponentButton::ExecuteInsertDeleteRule(const DriverInfo::TableInsertDeleteRule& inRule, int inSourceTableID, int inIndexPre, int inIndexPost)
	{

	}

	void ComponentButton::ExecuteAction(int inActionInput)
	{

	}
}