#include "component_check_button.h"

#include "foundation/graphics/textfield.h"
#include "foundation/input/mouse.h"
#include "foundation/input/keyboard.h"
#include "foundation/input/keyboard_utils.h"
#include "runtime/editor/cursor_control.h"

#include "SDL_keycode.h"
#include <assert.h>

using namespace Foundation;

namespace Editor
{
	ComponentCheckButton::ComponentCheckButton(int inID, int inGroupID, Undo* inUndo, Foundation::TextField* inTextField, bool inChecked, const std::string& inButtonText, int inX, int inY, int inWidth, std::function<void(int, bool)> inButtonPressedCallback)
		: ComponentBase(inID, inGroupID, inUndo, inTextField, inX, inY, inWidth + 4, 1)
		, m_ButtonText(inButtonText)
		, m_ButtonPressedCallback(inButtonPressedCallback)
		, m_MouseOver(false)
		, m_Checked(inChecked)
	{
		assert(inTextField != nullptr);
		assert(inButtonText.length() <= static_cast<size_t>(inWidth));
	}

	ComponentCheckButton::~ComponentCheckButton()
	{

	}

	//----------------------------------------------------------------------------------------------------------------------------------------

	void ComponentCheckButton::SetHasControl(GetControlType inGetControlType, CursorControl& inCursorControl)
	{
		m_HasControl = true;
		inCursorControl.SetEnabled(false);
		m_RequireRefresh = true;
	}


	bool ComponentCheckButton::ConsumeInput(const Foundation::Keyboard& inKeyboard, CursorControl& inCursorControl, ComponentsManager& inComponentsManager)
	{
		if (m_HasControl)
		{
			// Get key events
			for (const auto& key_event : inKeyboard.GetKeyEventList())
			{
				switch (key_event)
				{
				case SDLK_SPACE:
					DoChangeState();
					return true;
				default:
					break;
				}
			}
		}

		return false;
	}


	bool ComponentCheckButton::ConsumeInput(const Foundation::Mouse& inMouse, bool inModifierKeyMask, CursorControl& inCursorControl, ComponentsManager& inComponentsManager)
	{
		if(m_MouseOver)
		{
			if (inMouse.IsButtonPressed(Foundation::Mouse::Button::Left))
			{
				DoChangeState();
				return true;
			}
		}

		return false;
	}


	void ComponentCheckButton::ConsumeNonExclusiveInput(const Foundation::Mouse& inMouse)
	{
		Foundation::Point cell_position = m_TextField->GetCellPositionFromPixelPosition(inMouse.GetPosition());

		const bool mouse_over = (cell_position.m_Y == m_Position.m_Y && cell_position.m_X >= m_Position.m_X && cell_position.m_X < m_Position.m_X + m_Dimensions.m_Width);

		if (m_MouseOver != mouse_over)
			m_RequireRefresh = true;

		m_MouseOver = mouse_over;
	}

	//----------------------------------------------------------------------------------------------------------------------------------------

	void ComponentCheckButton::Refresh(const DisplayState& inDisplayState)
	{
		if (m_RequireRefresh && m_TextField->IsEnabled())
		{
			Color color = [&]()
			{
				if (m_MouseOver)
					return m_HasControl ? Color::LightRed : Color::LightRed;
				else
					return m_HasControl ? Color::Red : Color::DarkRed;
			}();
			m_TextField->ColorAreaBackground(color, m_Rect);
			m_TextField->Print(m_Position.m_X, m_Position.m_Y, m_Checked ? "[X]" : "[ ]");
			m_TextField->Print(m_Position.m_X + 4, m_Position.m_Y, m_ButtonText);
		}
	}


	void ComponentCheckButton::HandleDataChange()
	{
		if (m_HasDataChange)
			m_HasDataChange = false;
	}


	void ComponentCheckButton::PullDataFromSource()
	{

	}



	void ComponentCheckButton::ExecuteInsertDeleteRule(const DriverInfo::TableInsertDeleteRule& inRule, int inSourceTableID, int inIndexPre, int inIndexPost)
	{

	}

	void ComponentCheckButton::ExecuteAction(int inActionInput)
	{

	}


	void ComponentCheckButton::DoChangeState()
	{
		m_Checked = !m_Checked;
		m_ButtonPressedCallback(m_ComponentID, m_Checked);
	}

}