#include "dialog_utilities.h"

#include "foundation/graphics/viewport.h"
#include "foundation/graphics/textfield.h"
#include "foundation/graphics/types.h"
#include "foundation/graphics/color.h"
#include "foundation/input/keyboard.h"
#include "foundation/input/mouse.h"

#include "runtime/editor/components_manager.h"

using namespace Foundation;

namespace Editor
{
	DialogUtilities::DialogUtilities(int inWidth, int inHeight, std::function<void(const DialogUtilities::Selection)>&& inSelect, std::function<void(void)>&& inCancel)
		: DialogBase()
		, m_Width(inWidth)
		, m_Height(inHeight)
		, m_SelectFunction(inSelect)
		, m_CancelFunction(inCancel)
	{

	}


	void DialogUtilities::Cancel()
	{
		if (m_CancelFunction)
			m_CancelFunction();
	}


	bool DialogUtilities::ConsumeInput(const Foundation::Keyboard& inKeyboard, const Foundation::Mouse& inMouse)
	{
		DialogBase::ConsumeInput(inKeyboard, inMouse);

		for (const auto& key_event : inKeyboard.GetKeyEventList())
		{
			switch (key_event)
			{
			case SDLK_ESCAPE:
				m_Done = true;
				m_CancelFunction();

				return true;
			case SDLK_RETURN:
				{
					m_Done = true;
					const Selection selected = static_cast<Selection>(m_StringListSelectorComponent->GetSelectionIndex());
					m_SelectFunction(selected);
				}
				return true;
			}
		}

		return false;
	}


	void DialogUtilities::ActivateInternal(Foundation::Viewport* inViewport)
	{
		m_TextField = inViewport->CreateTextField(m_Width, m_Height, 0, 0);
		m_TextField->SetEnable(true);
		m_TextField->SetPositionToCenterOfViewport();

		m_TextField->ColorAreaBackground(Color::DarkRed);

		const std::string caption = "Utilities";

		m_TextField->ColorAreaBackground(Color::Red, { {0, 0}, {m_Width, 1} });
		m_TextField->Print((m_Width - static_cast<int>(caption.length())) >> 1, 0, caption);

		m_ComponentsManager = std::make_unique<ComponentsManager>(inViewport, m_CursorControl);
		m_ComponentsManager->SetGroupEnabledForInput(0, true);
		// m_ComponentsManager->SetGroupEnabledForTabbing(0);

		m_StringListDataBuffer = std::make_shared<DataSourceTList<std::string>>(std::vector<std::string>({ "Statistics", "Optimize", "Pack", "Clear Sequences", "Expand sequences" }));
		m_StringListSelectorComponent = std::make_shared<ComponentStringListSelector>
			(
				0, 0,
				nullptr,
				m_StringListDataBuffer,
				m_TextField,
				1, 1,
				m_Width - 2, m_Height,
				1, 1
			);

		m_StringListSelectorComponent->SetColors(Color::DarkRed, Color::DarkBlue, Color::DarkGrey);

		m_ComponentsManager->AddComponent(m_StringListSelectorComponent);
		m_ComponentsManager->SetComponentInFocus(m_StringListSelectorComponent);
	}


	void DialogUtilities::DeactivateInternal(Foundation::Viewport* inViewport)
	{
		inViewport->Destroy(m_TextField);
	}
}