#include "runtime/editor/editor_facility.h"
#include "runtime/editor/utilities/import_utils.h"
#include "runtime/environmentdefines.h"
#include "runtime/emulation/cpumos6510.h"
#include "runtime/emulation/cpumemory.h"
#include "runtime/emulation/sid/sidproxy.h"
#include "runtime/execution/executionhandler.h"
#include "runtime/execution/flightrecorder.h"
#include "runtime/editor/screens/screen_base.h"
#include "runtime/editor/screens/screen_intro.h"
#include "runtime/editor/screens/screen_edit.h"
#include "runtime/editor/screens/screen_disk.h"
#include "runtime/editor/screens/screen_edit_utils.h"
#include "runtime/editor/utilities/editor_utils.h"
#include "runtime/editor/auxilarydata/auxilary_data_collection.h"
#include "runtime/editor/auxilarydata/auxilary_data_hardware_preferences.h"
#include "runtime/editor/editor_types.h"
#include "runtime/editor/dialog/dialog_message.h"
#include "runtime/editor/dialog/dialog_message_yesno.h"
#include "runtime/editor/dialog/dialog_sid_file_info.h"
#include "runtime/editor/driver/driver_utils.h"
#include "runtime/editor/driver/driver_info.h"
#include "runtime/editor/packer/packer.h"
#include "foundation/graphics/viewport.h"
#include "foundation/graphics/textfield.h"
#include "foundation/platform/iplatform.h"
#include "foundation/input/keyboard.h"
#include "foundation/sound/audiostream.h"
#include "utils/utilities.h"
#include "utils/c64file.h"
#include "utils/psidfile.h"
#include "libraries/ghc/fs_std.h"

// System
#include <assert.h>

using namespace Foundation;
using namespace Emulation;
using namespace Utility;
using namespace fs;

namespace Editor
{
    const unsigned int EditorFacility::DefaultDialogWidth = 100;

	EditorFacility::EditorFacility(IPlatform* inPlatform, Viewport* inViewport, KeyHookStore& inKeyHookStore)
		: m_Viewport(inViewport)
		, m_Platform(inPlatform)
		, m_KeyHookStore(inKeyHookStore)
		, m_IsDone(false)
		, m_CurrentScreen(nullptr)
		, m_RequestedScreen(nullptr)
	{
		// Create emulation environment
		SIDConfiguration sid_configuration;										// Default settings are applicable

#ifdef _DEBUG
		sid_configuration.m_eSampleMethod = SID_SAMPLE_METHOD_FAST;				// Debug configuration is not fast enough to do any other kind of internal sampling!
#else
		sid_configuration.m_eSampleMethod = SID_SAMPLE_METHOD_RESAMPLE_INTERPOLATE;
#endif //_DEBUG
		sid_configuration.m_eModel = SID_MODEL_8580;

		m_SIDProxy = new SIDProxy(sid_configuration);
		m_CPUMemory = new CPUMemory(0x10000, m_Platform);
		m_CPU = new CPUmos6510();
		m_FlightRecorder = new FlightRecorder(m_Platform, 0x800);
		m_ExecutionHandler = new ExecutionHandler(m_Platform, m_CPU, m_CPUMemory, m_SIDProxy, m_FlightRecorder);

		// Create audio stream
		m_AudioStream = new AudioStream(44100, 16, 0x100, m_ExecutionHandler);

		// Create the main text field
		m_TextField = m_Viewport->CreateTextField(m_Viewport->GetWidth() / TextField::font_width, m_Viewport->GetHeight() / TextField::font_height, 0, 0);
		m_TextField->SetEnable(true);
		m_TextField->ColorAreaBackground(Color::Blue, 0, 0, 120, 1);

		// Allocate empty Driver Info class
		m_DriverInfo = std::make_shared<DriverInfo>();

		// Create screens
		m_IntroScreen = std::make_unique<ScreenIntro>(
			m_Viewport,
			m_TextField,
			&m_CursorControl,
			m_DisplayState,
			m_KeyHookStore,
			m_DriverInfo,
			[&]() { OnExitIntroScreen(); },
			[&]() { OnExitIntroScreenToLoad(); }
		);

		m_DiskScreen = std::make_unique<ScreenDisk>(
			m_Platform,
			m_Viewport,
			m_TextField,
			&m_CursorControl,
			m_DisplayState,
			m_KeyHookStore,
			[&](const std::string& inFilenameSelection, FileType inSaveFileType) { OnFilenameSelection(m_DiskScreen.get(), inFilenameSelection, inSaveFileType); },
			[&]() { OnCancelScreen(m_DiskScreen.get()); });

		m_EditScreen = std::make_unique<ScreenEdit>(
			m_Viewport,
			m_TextField,
			&m_CursorControl,
			m_DisplayState,
			m_KeyHookStore,
			m_EditState,
			m_CPUMemory,
			m_ExecutionHandler,
			m_SIDProxy,
			m_DriverInfo,
			[&]() {	m_DiskScreen->SetMode(ScreenDisk::Load); RequestScreen(m_DiskScreen.get()); },
			[&]() {	m_DiskScreen->SetMode(ScreenDisk::Save); m_DiskScreen->SetSuggestedFileName(m_LastSF2PathAndFilename);  RequestScreen(m_DiskScreen.get()); },
			[&]() { m_DiskScreen->SetMode(ScreenDisk::Import); RequestScreen(m_DiskScreen.get()); },
			[&]() {	m_DiskScreen->SetMode(ScreenDisk::LoadInstrument); RequestScreen(m_DiskScreen.get()); },
			[&]() {	m_DiskScreen->SetMode(ScreenDisk::SaveInstrument); m_DiskScreen->SetSuggestedFileName(m_LastSF2PathAndFilename);  RequestScreen(m_DiskScreen.get()); },
			[&]() { OnQuickSave(m_EditScreen.get()); },
			[&](unsigned short inDestinationAddress) { OnPack(m_EditScreen.get(), inDestinationAddress); }
		);
	}

	EditorFacility::~EditorFacility()
	{
		m_AudioStream->Stop();

		m_Viewport->Destroy(m_TextField);

		delete m_AudioStream;
		delete m_ExecutionHandler;
		delete m_FlightRecorder;
		delete m_SIDProxy;
		delete m_CPU;
		delete m_CPUMemory;
	}

	//--------------------------------------------------------------------------------

	void EditorFacility::Start(const char* inFileToLoad)
	{
		assert(m_ExecutionHandler != nullptr);
		assert(m_AudioStream != nullptr);

		bool file_loaded_successfully = [&]()
		{
			if (inFileToLoad != nullptr)
			{
				std::string file_to_load(inFileToLoad);
				return LoadFile(file_to_load);
			}

			return false;
		}();

		// Try to load the driver directly
		if(!file_loaded_successfully)
            LoadFile("drivers//sf2driver11_00.prg");

        // After loading, set the current path, so that opening the disk menu will be correct.
        std::error_code ec;
        fs::current_path(m_Platform->Storage_GetStartupPath(), ec);
        
        // Start the intro screen
		SetCurrentScreen(m_IntroScreen.get());

		m_AudioStream->Start();
		m_ExecutionHandler->Start();
	}

	void EditorFacility::Stop()
	{
		assert(m_AudioStream != nullptr);

		m_AudioStream->Stop();
	}

	//--------------------------------------------------------------------------------

	void EditorFacility::Update(const Keyboard& inKeyboard, const Mouse& inMouse, int inDeltaTicks)
	{
		if (m_IsDone)
			return;

		// Check screen status
		HandleScreenState();

		// Handle component updates
		if (m_CurrentScreen != nullptr)
		{
			m_CurrentScreen->ConsumeInput(inKeyboard, inMouse);
			m_CurrentScreen->Update(inDeltaTicks);
		}

		// Update cursor control
		m_CursorControl.Update(inDeltaTicks);

		// Handle viewport updates
		m_Viewport->Begin();

		if (m_CurrentScreen != nullptr)
			m_CurrentScreen->Refresh();

		m_Viewport->End();
	}

	//------------------------------------------------------------------------------------------------------------

	bool EditorFacility::IsDone() const
	{
		return m_IsDone;
	}

	void EditorFacility::TryQuit()
	{
		if (m_CurrentScreen != nullptr)
		{
			m_CurrentScreen->TryQuit([&](bool inQuit)
			{
				if (inQuit)
					m_IsDone = true;
			});
		}
	}

	void EditorFacility::TryLoad(const std::string inPathAndFilename)
	{
		if (m_CurrentScreen != nullptr)
		{
			m_CurrentScreen->TryLoad(inPathAndFilename, [&, path_and_filename = inPathAndFilename](bool inQuit)
			{
				if (LoadFile(path_and_filename))
					ForceRequestScreen(m_EditScreen.get());
				else
					m_CurrentScreen->GetComponentsManager().StartDialog(std::make_shared<DialogMessage>("Invalid file", "The selected file is not compatible with SID Factory II.", DefaultDialogWidth, true, []() {}));
			});
		}
	}


	//------------------------------------------------------------------------------------------------------------

	void EditorFacility::RequestScreen(ScreenBase* inRequestedScreen)
	{
		assert(m_RequestedScreen == nullptr);

		if(m_CurrentScreen != inRequestedScreen)
			m_RequestedScreen = inRequestedScreen;
	}

	void EditorFacility::ForceRequestScreen(ScreenBase* inRequestedScreen)
	{
		assert(m_RequestedScreen == nullptr);
		m_RequestedScreen = inRequestedScreen;
	}

	void EditorFacility::SetCurrentScreen(ScreenBase* inCurrentScreen)
	{
		if (m_CurrentScreen != nullptr)
			m_CurrentScreen->Deactivate();

		m_CurrentScreen = inCurrentScreen;

		if (m_CurrentScreen != nullptr)
			m_CurrentScreen->Activate();
	}

	void EditorFacility::HandleScreenState()
	{
		if (m_RequestedScreen != nullptr)
		{
			SetCurrentScreen(m_RequestedScreen);
			m_RequestedScreen = nullptr;
		}
	}

	//------------------------------------------------------------------------------------------------------------

	bool EditorFacility::IsFileSF2(const std::string& inPathAndFilename)
	{
		// Read test music data to cpu memory
		void* data = nullptr;
		long data_size = 0;

		std::shared_ptr<DriverInfo> driver_info = std::make_shared<DriverInfo>();

		if (Utility::ReadFile(inPathAndFilename, 65536, &data, data_size))
		{
			if (data_size > 2)
			{
				std::shared_ptr<Utility::C64File> c64_file = Utility::C64File::CreateFromPRGData(data, static_cast<unsigned int>(data_size));

				if(c64_file != nullptr)
					driver_info->Parse(*c64_file);
			}
		}

		return driver_info->IsValid() || driver_info->IsParticalyValid();
	}


	bool EditorFacility::LoadFile(const std::string& inPathAndFilename)
	{
		// Read test music data to cpu memory
		void* data = nullptr;
		long data_size = 0;

		std::shared_ptr<DriverInfo> driver_info = std::make_shared<DriverInfo>();

		if (Utility::ReadFile(inPathAndFilename, 65536, &data, data_size))
		{
			if (data_size > 2)
			{
				std::shared_ptr<Utility::C64File> c64_file = Utility::C64File::CreateFromPRGData(data, static_cast<unsigned int>(data_size));

				if (c64_file != nullptr)
				{
					driver_info->Parse(*c64_file);

					if (driver_info->IsValid())
					{
						m_DriverInfo = driver_info;
						m_DriverInfo->GetAuxilaryDataCollection().Reset();

						const unsigned short auxilary_data_vector = m_DriverInfo->GetDriverCommon().m_InitAddress - 5;
						const unsigned short auxilary_data_address = c64_file->GetWord(auxilary_data_vector);

						if (auxilary_data_address != 0)
						{
							Utility::C64FileReader reader = Utility::C64FileReader(*c64_file, auxilary_data_address);
							m_DriverInfo->GetAuxilaryDataCollection().Load(reader);
						}

						// Copy the data to the emulated memory
						m_CPUMemory->Lock();
						m_CPUMemory->Clear();
						m_CPUMemory->SetData(c64_file->GetTopAddress(), c64_file->GetData(), c64_file->GetDataSize());
						m_CPUMemory->Unlock();

						// Init the execution handler 
						m_ExecutionHandler->SetInitVector(m_DriverInfo->GetDriverCommon().m_InitAddress);
						m_ExecutionHandler->SetStopVector(m_DriverInfo->GetDriverCommon().m_StopAddress);
						m_ExecutionHandler->SetUpdateVector(m_DriverInfo->GetDriverCommon().m_UpdateAddress);

						// Store name of last read file
						SetLastSavedPathAndFilename(inPathAndFilename);

						// Flush undo after load
						m_EditScreen->FlushUndo();
					}
				}
			}

			delete[] static_cast<char*>(data);
		}

		return driver_info->IsValid();
	}


	bool EditorFacility::SaveFile(const std::string& inPathAndFilename)
	{
		if (m_DriverInfo->IsValid())
		{
			m_CPUMemory->Lock();

			const unsigned short top_of_file_address = m_DriverInfo->GetTopAddress();
			const unsigned short end_of_file_address = DriverUtils::GetEndOfMusicDataAddress(*m_DriverInfo, reinterpret_cast<const Emulation::IMemoryRandomReadAccess&>(*m_CPUMemory));
			const unsigned short data_size = end_of_file_address - top_of_file_address;

			unsigned char* data = new unsigned char[data_size];
			m_CPUMemory->GetData(top_of_file_address, data, data_size);
			m_CPUMemory->Unlock();
			std::shared_ptr<Utility::C64File> file = Utility::C64File::CreateFromData(top_of_file_address, data, data_size);
			delete[] data;

			Utility::C64FileWriter file_writer(*file, end_of_file_address, true);

			const unsigned short irq_vector = file_writer.GetWriteAddress();
			DriverUtils::InsertIRQ(*m_DriverInfo, file_writer);

			const unsigned short auxilary_data_vector = file_writer.GetWriteAddress();
			m_DriverInfo->GetAuxilaryDataCollection().Save(file_writer);

			// Adjust IRQ and auxilary data vectors in file
			const unsigned short driver_init_vector = m_DriverInfo->GetDriverCommon().m_InitAddress;
			(*file)[driver_init_vector - 2] = static_cast<unsigned char>(irq_vector & 0xff);
			(*file)[driver_init_vector - 1] = static_cast<unsigned char>(irq_vector >> 8);
			(*file)[driver_init_vector - 5] = static_cast<unsigned char>(auxilary_data_vector & 0xff);
			(*file)[driver_init_vector - 4] = static_cast<unsigned char>(auxilary_data_vector >> 8);

			// Save to disk
			if(!Utility::WriteFile(inPathAndFilename, file))
                return false;

			SetLastSavedPathAndFilename(inPathAndFilename);

			return true;
		}

		return false;
	}


	bool EditorFacility::LoadFileForImport(const std::string& inPathAndFilename, std::shared_ptr<DriverInfo>& outDriverInfo, std::shared_ptr<Utility::C64File>& outC64File)
	{
		// Read test music data to cpu memory
		void* data = nullptr;
		long data_size = 0;

		outDriverInfo = std::make_shared<DriverInfo>();

		if (Utility::ReadFile(inPathAndFilename, 65536, &data, data_size))
		{
			if (data_size > 2)
			{
				outC64File = Utility::C64File::CreateFromPRGData(data, static_cast<unsigned int>(data_size));
				if (outC64File != nullptr)
				{
					outDriverInfo->Parse(*outC64File);

					// Store name of last read file
					SetLastSavedPathAndFilename(inPathAndFilename);

					// Flush undo after load
					m_EditScreen->FlushUndo();

					return true;
				}
			}

			delete[] static_cast<char*>(data);
		}

		return false;
	}



	bool EditorFacility::SavePackedFile(const std::string& inFileName)
	{
		if (m_PackedData != nullptr)
		{
			// Create file
			std::shared_ptr<Utility::C64File> file = Utility::C64File::CreateFromData(m_PackedData->GetTopAddress(), m_PackedData->GetData(), m_PackedData->GetDataSize());

			// Save to disk
			Utility::WriteFile(inFileName, file);

			return true;
		}

		return false;
	}


	bool EditorFacility::SavePackedFileToSID(ScreenBase* inCallerScreen, const std::string& inFileName)
	{
		if (m_PackedData != nullptr)
		{
			auto do_save = [&, inFileName](std::string inTitle, std::string inAuthor, std::string inCopyright)
			{
				unsigned short top_of_file_address = m_PackedData->GetTopAddress();
				unsigned short data_size = m_PackedData->GetDataSize();

				unsigned char* data = new unsigned char[data_size + 2];

				data[0] = static_cast<unsigned char>(top_of_file_address & 0xff);
				data[1] = static_cast<unsigned char>(top_of_file_address >> 8);

				unsigned char* packed_data = m_PackedData->GetData();

				for (int i = 0; i < data_size; ++i)
					data[i + 2] = packed_data[i];

				// Save PSID file to disk, also
				const auto& driver_common = m_DriverInfo->GetDriverCommon();
				const auto& hardware_preferences = m_DriverInfo->GetAuxilaryDataCollection().GetHardwarePreferences();

				Utility::PSIDFile psid_file(
					data,
					data_size + 2,
					0,
					driver_common.m_UpdateAddress - driver_common.m_InitAddress,
					1,
					inTitle,
					inAuthor,
					inCopyright,
					hardware_preferences.GetSIDModel() == AuxilaryDataHardwarePreferences::MOS6581,
					hardware_preferences.GetRegion() == AuxilaryDataHardwarePreferences::PAL);

				const unsigned char* psid_data = psid_file.GetData();

				Utility::WriteFile(inFileName, psid_data, psid_file.GetDataSize());

				delete[] data;

				RequestScreen(m_EditScreen.get());
			};

			inCallerScreen->GetComponentsManager().StartDialog(std::make_shared<DialogSIDFileInfo>(do_save, []() {}));

			return true;
		}

		return false;
	}
	//-------------------------------------------------------------------------------------------------------------------------

	void EditorFacility::OnCancelScreen(ScreenBase* inCallerScreen)
	{
		if (inCallerScreen == m_DiskScreen.get())
		{
			if (m_DriverInfo->IsValid())
				RequestScreen(m_EditScreen.get());
			else
				inCallerScreen->GetComponentsManager().StartDialog(std::make_shared<DialogMessage>("No driver loaded", "Cannot enter the editor when no driver has been loaded. Please load a valid file from the file selection screen!", DefaultDialogWidth, true, []() {}));
		}
	}


	void EditorFacility::OnFilenameSelection(ScreenBase* inCallerScreen, const std::string& inSelectedFilename, FileType inFileType)
	{
		if (inCallerScreen == m_DiskScreen.get())
		{
			// Handle loading
			if (m_DiskScreen->GetMode() == ScreenDisk::Load)
			{
				assert(inFileType == FileType::SF2);
				DoLoad(inCallerScreen, inSelectedFilename);
			}

			// Handle saving
			if (m_DiskScreen->GetMode() == ScreenDisk::Save)
			{
				assert(inFileType == FileType::SF2);
				DoSave(inCallerScreen, inSelectedFilename);
			}

			// Handle importing
			if (m_DiskScreen->GetMode() == ScreenDisk::Import)
			{
				assert(inFileType == FileType::SF2);
				DoImport(inCallerScreen, inSelectedFilename);
			}

			// Handle loading
			if (m_DiskScreen->GetMode() == ScreenDisk::LoadInstrument)
			{
				assert(inFileType == FileType::SI2);
				DoLoadInstrument(inCallerScreen, inSelectedFilename);
			}

			// Handle saving
			if (m_DiskScreen->GetMode() == ScreenDisk::SaveInstrument)
			{
				assert(inFileType == FileType::SI2);
				DoSaveInstrument(inCallerScreen, inSelectedFilename);
			}

			// Handle save packed
			if (m_DiskScreen->GetMode() == ScreenDisk::SavePacked)
			{
				if (inFileType == FileType::PRG)
					DoSavePacked(inCallerScreen, inSelectedFilename);
				else
				{
					assert(inFileType == FileType::SID);
					DoSavePackedToSID(inCallerScreen, inSelectedFilename);
				}
			}
		}
	}


	void EditorFacility::OnExitIntroScreen()
	{
		assert(m_DriverInfo != nullptr);

		if (m_DriverInfo->IsValid())
			SetCurrentScreen(m_EditScreen.get());
		else
		{
			// .. otherwise go to the load screen
			m_DiskScreen->SetMode(ScreenDisk::Mode::Load);
			SetCurrentScreen(m_DiskScreen.get());
		}
	}


	void EditorFacility::OnExitIntroScreenToLoad()
	{
		m_DiskScreen->SetMode(ScreenDisk::Mode::Load);
		SetCurrentScreen(m_DiskScreen.get());
	}


	void EditorFacility::OnPack(ScreenBase* inCallerScreen, unsigned short inDestinationAdress)
	{
		const bool is_uppercase = m_DisplayState.IsHexUppercase();

		Packer packer(*m_CPUMemory, *m_DriverInfo, inDestinationAdress);
		m_PackedData = packer.GetResult();

		std::string packing_info;
		packing_info += "Range: 0x" + EditorUtils::ConvertToHexValue(static_cast<unsigned short>(m_PackedData->GetTopAddress()), is_uppercase) + " - 0x" + EditorUtils::ConvertToHexValue(static_cast<unsigned short>(m_PackedData->GetBottomAddress()), is_uppercase) + "\n";
		packing_info += "Size : 0x" + EditorUtils::ConvertToHexValue(static_cast<unsigned short>(m_PackedData->GetDataSize()), is_uppercase);

		inCallerScreen->GetComponentsManager().StartDialog(
			std::make_shared<DialogMessage>("Packing results", packing_info, 30, false, [&]() 
			{
				m_DiskScreen->SetMode(ScreenDisk::Mode::SavePacked);
				SetCurrentScreen(m_DiskScreen.get());
			})
		);
	}


	void EditorFacility::OnQuickSave(ScreenBase* inCallerScreen)
	{
		DoQuickSave(inCallerScreen);
	}


    void EditorFacility::OnSaveError(ScreenBase* inCallerScreen)
    {
        inCallerScreen->GetComponentsManager().StartDialog(std::make_shared<DialogMessage>("Error", "The file could not be saved to the current destination!", DefaultDialogWidth, true, []() {}));
    }

	//-------------------------------------------------------------------------------------------------------------------------


	void EditorFacility::DoLoad(ScreenBase* inCallerScreen, const std::string& inSelectedFilename)
	{
		auto do_load = [&, inSelectedFilename, inCallerScreen]()
		{
			if (LoadFile(inSelectedFilename))
				RequestScreen(m_EditScreen.get());
			else
				inCallerScreen->GetComponentsManager().StartDialog(std::make_shared<DialogMessage>("Invalid file", "The selected file is not compatible with SID Factory II.", DefaultDialogWidth, true, []() {}));
		};

		if (m_DriverInfo->IsValid())
			inCallerScreen->GetComponentsManager().StartDialog(std::make_shared<DialogMessageYesNo>("Load file", "Are you sure you want to load:\n" + inSelectedFilename + "? \nAny unsaved changes will be lost!", DefaultDialogWidth, do_load, []() {}));
		else
			do_load();
	}


	void EditorFacility::DoSave(ScreenBase* inCallerScreen, const std::string& inSelectedFilename)
	{
		path save_path_and_filename = inSelectedFilename;

		if (!exists(save_path_and_filename))
		{
			if(SaveFile(save_path_and_filename.string()))
                RequestScreen(m_EditScreen.get());
            else
                OnSaveError(inCallerScreen);
		}
		else if (IsFileSF2(save_path_and_filename.string()))
		{
			auto do_save = [save_path_and_filename, inCallerScreen, this]()
			{
				if(SaveFile(save_path_and_filename.string()))
                    this->RequestScreen(this->m_EditScreen.get());
                else
                    OnSaveError(inCallerScreen);
			};

			inCallerScreen->GetComponentsManager().StartDialog(std::make_shared<DialogMessageYesNo>("Warning", inSelectedFilename + "\nAlready exists! Are you sure you want to overwrite it?", DefaultDialogWidth, do_save, []() {}));
		}
		else
		{
			inCallerScreen->GetComponentsManager().StartDialog(std::make_shared<DialogMessage>("Illegal save destination", "You are trying to overwrite a file, which cannot be idenitified as belonging to SID Factory II.\nPlease choose another name!", DefaultDialogWidth, true, []() {}));
		}
	}


	void EditorFacility::DoLoadInstrument(ScreenBase* inCallerScreen, const std::string& inSelectedFilename)
	{
		RequestScreen(m_EditScreen.get());
	}


	void EditorFacility::DoSaveInstrument(ScreenBase* inCallerScreen, const std::string& inSelectedFilename)
	{
		RequestScreen(m_EditScreen.get());
	}


	void EditorFacility::DoQuickSave(ScreenBase* inCallerScreen)
	{
		path save_path_and_filename = m_LastSF2PathAndFilename;

		const bool sf2_extension = save_path_and_filename.extension().string() == ".sf2";

		if (!exists(save_path_and_filename))
			inCallerScreen->GetComponentsManager().StartDialog(std::make_shared<DialogMessage>("Illegal save destination", "The quick save file does not exist!", DefaultDialogWidth, true, []() {}));
		else if (!IsFileSF2(save_path_and_filename.string()))
			inCallerScreen->GetComponentsManager().StartDialog(std::make_shared<DialogMessage>("Illegal save destination", "You are trying to quick save to a file, which cannot be idenitified as belonging to SID Factory II.\nPlease save through the save disk menu!", DefaultDialogWidth, true, []() {}));
		else if (!sf2_extension)
			inCallerScreen->GetComponentsManager().StartDialog(std::make_shared<DialogMessage>("Illegal save destination", "You are trying to quick save to a file, with an extension other than .sf2.\nPlease save through the save disk menu!", DefaultDialogWidth, true, []() {}));
		else
		{
			
			auto do_save = [save_path_and_filename, inCallerScreen, this]()
			{
				if(SaveFile(save_path_and_filename.string()))
                    this->m_EditScreen->SetStatusBarMessage(" Quick saved to: " + save_path_and_filename.filename().string(), 5000);
                else
                    this->OnSaveError(inCallerScreen);
			};

			inCallerScreen->GetComponentsManager().StartDialog(std::make_shared<DialogMessageYesNo>("Warning", "Do you want to perform a quick save to:\n" + save_path_and_filename.string() + "?", DefaultDialogWidth, do_save, []() {}));
		}
	}


	void EditorFacility::DoImport(ScreenBase* inCallerScreen, const std::string& inSelectedFilename)
	{
		auto do_load = [&, inSelectedFilename, inCallerScreen]()
		{
			std::shared_ptr<DriverInfo> import_driver_info = nullptr;
			std::shared_ptr<Utility::C64File> import_c64_file = nullptr;

			if (LoadFileForImport(inSelectedFilename, import_driver_info, import_c64_file))
			{
				const bool valid_tables = import_driver_info->HasParsedHeaderBlock(DriverInfo::HeaderBlockID::ID_DriverTables);
				const bool valid_music_data = import_driver_info->HasParsedHeaderBlock(DriverInfo::HeaderBlockID::ID_MusicData);

				if (!valid_music_data)
					inCallerScreen->GetComponentsManager().StartDialog(std::make_shared<DialogMessage>("Import failure", "Unable to import music data!", DefaultDialogWidth, true, []() {}));
				else if(!valid_tables)
					inCallerScreen->GetComponentsManager().StartDialog(std::make_shared<DialogMessage>("Import failure", "Unable to import table definitions!", DefaultDialogWidth, true, []() {}));
				else 
				{
					ImportUtils::Import(inCallerScreen, *m_DriverInfo, *m_CPUMemory, *import_driver_info, *import_c64_file);
					this->RequestScreen(this->m_EditScreen.get());
				}

			}
			else
				inCallerScreen->GetComponentsManager().StartDialog(std::make_shared<DialogMessage>("Invalid file", "The selected file is not compatible with SID Factory II import.", DefaultDialogWidth, true, []() {}));
		};

		inCallerScreen->GetComponentsManager().StartDialog(std::make_shared<DialogMessageYesNo>("Import song data", "Are you sure you want to import the song data from\n" + inSelectedFilename + "?\nAny unsaved changes will be lost!", DefaultDialogWidth, do_load, []() {}));
	}


	void EditorFacility::DoSavePacked(ScreenBase* inCallerScreen, const std::string& inSelectedFilename)
	{
		path save_path_and_filename = inSelectedFilename;

		if (!exists(save_path_and_filename))
		{
			SavePackedFile(save_path_and_filename.string());
			RequestScreen(m_EditScreen.get());
		}
		else
		{
			auto do_save = [save_path_and_filename, this]()
			{
				SavePackedFile(save_path_and_filename.string());
				this->RequestScreen(this->m_EditScreen.get());
			};

			inCallerScreen->GetComponentsManager().StartDialog(std::make_shared<DialogMessageYesNo>("Warning", inSelectedFilename + "\nAlready exists! Are you sure you want to overwrite it?", DefaultDialogWidth, do_save, []() {}));
		}
	}


	void EditorFacility::DoSavePackedToSID(ScreenBase* inCallerScreen, const std::string& inSelectedFilename)
	{
		path save_path_and_filename = inSelectedFilename;

		if (!exists(save_path_and_filename))
		{
			SavePackedFileToSID(inCallerScreen, save_path_and_filename.string());
		}
		else
		{
			auto do_save = [inCallerScreen, save_path_and_filename, this]()
			{
				SavePackedFileToSID(inCallerScreen, save_path_and_filename.string());
			};

			inCallerScreen->GetComponentsManager().StartDialog(std::make_shared<DialogMessageYesNo>("Warning", inSelectedFilename + "\nAlready exists! Are you sure you want to overwrite it?", DefaultDialogWidth, do_save, []() {}));
		}
	}


	void EditorFacility::SetLastSavedPathAndFilename(const std::string& inLastSavedPathAndFilename)
	{
		m_LastSF2PathAndFilename = inLastSavedPathAndFilename;

		const size_t length = inLastSavedPathAndFilename.size();

		for (size_t i = length - 1; i >= 0; --i)
		{
			const char character = inLastSavedPathAndFilename[i];
			if (character == '/' || character == '\\')
			{
				std::string file_name = inLastSavedPathAndFilename.substr(i + 1, length - (i + 1));
				m_Viewport->SetAdditionTitleInfo(file_name);

				return;
			}
		}

		m_Viewport->SetAdditionTitleInfo("");
	}

}

