#include "sidproxy.h"
#include "runtime/environmentdefines.h"

#ifdef __USE_SIDFP__
#include "libraries/resid-fp/sidfp.h"
#else
#include "libraries/resid/sid.h"
#endif //__USE_SIDFP__

#include <assert.h>
#include <cmath>

namespace Emulation
{
	SIDProxy::SIDProxy(const SIDConfiguration& sConfiguration)
		: m_sConfiguration(sConfiguration)
		, m_SampleCounter(0)
	{
		// Create instance of reSid
#ifdef __USE_SIDFP__
		m_pSID = new SIDFP();
#else
		m_pSID = new SID();
#endif //__USE_SIDFP__

		// Apply settings
		ApplySettings();
	}

	SIDProxy::~SIDProxy()
	{
		delete m_pSID;
	}

	//------------------------------------------------------------------------------------------------------------

	void SIDProxy::SetEnvironment(SIDEnvironment eEnvironment)
	{
		m_sConfiguration.m_eEnvironment = eEnvironment;
	}


	void SIDProxy::SetModel(SIDModel eModel)
	{
		m_sConfiguration.m_eModel = eModel;
	}


	void SIDProxy::SetSampleMethod(SIDSampleMethod eSampleMethod)
	{
		m_sConfiguration.m_eSampleMethod = eSampleMethod;
	}


	void SIDProxy::SetSampleFrequency(int nSampleFrequency)
	{
		m_sConfiguration.m_nSampleFrequency = nSampleFrequency;
	}


	SIDEnvironment SIDProxy::GetEnvironment() const
	{
		return m_sConfiguration.m_eEnvironment;
	}


	SIDModel SIDProxy::GetModel() const
	{
		return m_sConfiguration.m_eModel;
	}


	SIDSampleMethod SIDProxy::GetSampleMethod() const
	{
		return m_sConfiguration.m_eSampleMethod;
	}


	int SIDProxy::GetSampleFrequency() const
	{
		return m_sConfiguration.m_nSampleFrequency;
	}

	//------------------------------------------------------------------------------------------------------------

	void SIDProxy::SetConfiguration(const SIDConfiguration& sConfiguration)
	{
		// Copy the configuration
		m_sConfiguration = sConfiguration;

		// Apply settings of configuration to sid emulation layer
		ApplySettings();
	}

	//------------------------------------------------------------------------------------------------------------

	void SIDProxy::ApplySettings()
	{
		if (m_pSID != nullptr)
		{
			// Reset the sid
			m_pSID->reset();

			// Convert sampling methods flag
			sampling_method eSamplingMethod = SAMPLE_FAST;

			switch (m_sConfiguration.m_eSampleMethod)
			{
			case SID_SAMPLE_METHOD_FAST:
				break;
			case SID_SAMPLE_METHOD_INTERPOLATE:
				eSamplingMethod = SAMPLE_INTERPOLATE;
				break;
			case SID_SAMPLE_METHOD_RESAMPLE_FAST:
				eSamplingMethod = SAMPLE_RESAMPLE_FAST;
				break;
			case SID_SAMPLE_METHOD_RESAMPLE_INTERPOLATE:
				eSamplingMethod = SAMPLE_RESAMPLE_INTERPOLATE;
				break;
			}

			// Set sampling parameters
			double passband = 20000.0;

			m_pSID->set_sampling_parameters(
				static_cast<double>(m_sConfiguration.m_eEnvironment == SID_ENVIRONMENT_PAL ? EMULATION_CYCLES_PER_SECOND_PAL : EMULATION_CYCLES_PER_SECOND_NTSC),
				eSamplingMethod,
				static_cast<double>(m_sConfiguration.m_nSampleFrequency),
				passband);

			// Set chip model to emulate
			m_pSID->set_chip_model(m_sConfiguration.m_eModel == SID_MODEL_6581 ? MOS6581 : MOS8580);

			// Set defaults
			m_pSID->input(0);
			m_pSID->enable_filter(true);

#ifdef __USE_SIDFP__
			struct FILTERPARAMS
			{
				float distortionrate;
				float distortionpoint;
				float distortioncfthreshold;
				float type3baseresistance;
				float type3offset;
				float type3steepness;
				float type3minimumfetresistance;
				float type4k;
				float type4b;
				float voicenonlinearity;
			} filterparams =
			{ 
				0.50f, 3.3e6f, 1.0e-4f,
				1147036.4394268463f, 274228796.97550374f, 1.0066634233403395f, 16125.154840564108f,
				5.5f, 20.f,
				0.9613160610660189f 
			};

			m_pSID->get_filter().set_distortion_properties(
				filterparams.distortionrate,
				filterparams.distortionpoint,
				filterparams.distortioncfthreshold);
			m_pSID->get_filter().set_type3_properties(
				filterparams.type3baseresistance,
				filterparams.type3offset,
				filterparams.type3steepness,
				filterparams.type3minimumfetresistance);
			m_pSID->get_filter().set_type4_properties(
				filterparams.type4k,
				filterparams.type4b);
			m_pSID->set_voice_nonlinearity(
				filterparams.voicenonlinearity);
#else
			m_pSID->enable_external_filter(false);
#endif

		}
	}

	//------------------------------------------------------------------------------------------------------------

	void SIDProxy::StartRecordToFile(const std::string& inFileName)
	{
		assert(m_FileName.empty());
		m_FileName = inFileName;
	}

	void SIDProxy::StopRecordToFile()
	{
		assert(!m_FileName.empty());

		if (m_FileOutput.size() > 0)
		{
			FILE* file = fopen(m_FileName.c_str(), "wb");

			struct wave_header
			{
				char chunk_id[4] = { 'R', 'I', 'F', 'F' };
				int chunk_size;
				char format[4] = { 'W', 'A', 'V', 'E' }; 
				char sub_chunk1_id[4] = { 'f', 'm', 't', ' ' };
				int sub_chunk1_size;
				short audio_format = 1;
				short num_channels = 1;			//    Mono = 1, Stereo = 2, etc.
				int sample_rate = 44100;		//      8000, 44100, etc.
				int byte_rate = 88200;			//== SampleRate * NumChannels * BitsPerSample / 8
				short block_align = 2;			// == NumChannels * BitsPerSample / 8
				short bits_per_sample = 16;		//    8 bits = 8, 16 bits = 16, etc.
				char sub_chunk2_id[4] = { 'd', 'a', 't', 'a' };
				int sub_chunk2_size;
			} header;

			const int data_size = static_cast<int>(m_FileOutput.size()) * 2;
			header.chunk_size = 36 + data_size;
			header.sub_chunk1_size = 16;
			header.sub_chunk2_size = data_size;

			fwrite(&header, sizeof(wave_header), 1, file);

			for (short sample : m_FileOutput)
				fwrite(&sample, 2, 1, file);

			fclose(file);
		}

		m_FileOutput.clear();
		m_FileName.clear();
	}

	bool SIDProxy::IsRecordingToFile() const
	{
		return !m_FileName.empty();
	}

	//------------------------------------------------------------------------------------------------------------

	void SIDProxy::Reset()
	{
		assert(m_pSID != nullptr);

		m_pSID->reset();
	}

	int SIDProxy::Clock(int& nDeltaCycles, short* pBuffer, int nBufferSize)
	{
		assert(m_pSID != nullptr);

		// Cast to reSid type
		cycle_count nInternalDeltaCycles = static_cast<cycle_count>(nDeltaCycles);

		// Clock
		int nSamplesWritten = m_pSID->clock(nInternalDeltaCycles, pBuffer, nBufferSize);

		// Overwrite with sine wave to test output consistency
//		for (int i = 0; i < nSamplesWritten; ++i)
//		{
//			float r = (static_cast<float>(m_SampleCounter) * 2.0f * 3.1416f) / 25.0f;
//			short v = static_cast<short>(std::sinf(r) * 65535.0f / 4.0f);
//		
//			pBuffer[i] = v;
//		
//			m_SampleCounter++;
//		}

		if (IsRecordingToFile())
		{
			m_FileOutput.push_back(0);
			for (int i = 0; i < nSamplesWritten; ++i)
				m_FileOutput.push_back(pBuffer[i]);
		}

		// Cast back to int
		nDeltaCycles = static_cast<int>(nInternalDeltaCycles);

		// Return number of samples written!
		return nSamplesWritten;
	}

	void SIDProxy::Write(unsigned char ucReg, unsigned char ucValue)
	{
		assert(m_pSID != nullptr);

		m_pSID->write(static_cast<reg8>(ucReg), static_cast<reg8>(ucValue));
	}
}
