#pragma once

#include "SDL_keycode.h"
#include "foundation/input/keyboard_utils.h"
#include "utils/keyhookstore.h"
#include <functional>
#include <vector>
#include <string>

namespace Utility
{
	template<typename CONTEXT>
	class KeyHook
	{
	public:
		KeyHook();
		KeyHook(const std::string& inIdentifier, const KeyHookStore& inKeyHookStore, std::function<CONTEXT> inKeyDownCallback);
		KeyHook(const std::string& inIdentifier, SDL_Keycode inKeyCode, unsigned int inModifier, std::function<CONTEXT> inKeyDownCallback);

		const bool operator == (KeyHook<CONTEXT>& inOther) const;
		const bool operator != (KeyHook<CONTEXT>& inOther) const;

		operator bool() const;

		const std::string& GetIdentifier() const;

		void ConfigureKey(SDL_Keycode inKeyCode, unsigned int inModifier);

		bool TryConsume(SDL_Keycode inKeyCode, unsigned int inModifier) const;

		template<typename ...ARGS>
		bool TryConsume(SDL_Keycode inKeyCode, unsigned int inModifier, ARGS... inArgs) const;

	private:
		bool RequireShiftDown() const;
		bool RequireControlDown() const;
		bool RequireAltDown() const;

		template<typename ...ARGS>
		bool OnConsume(ARGS... inArgs) const;

		std::function<CONTEXT> m_KeyDownCallback;

		std::string m_Identifier;
		SDL_Keycode m_KeyCode;
		unsigned int m_Modifiers;
	};


	template<typename CONTEXT>
	inline bool ConsumeInputKeyHooks(SDL_Keycode inKeyCode, unsigned int inModifier, const std::vector<KeyHook<CONTEXT>>& inKeyHookList)
	{
		for (const auto& key_hook : inKeyHookList)
		{
			if (key_hook.TryConsume(inKeyCode, inModifier))
				return true;
		}
		return false;
	}


	template<typename CONTEXT, typename ...ARGS>
	inline bool ConsumeInputKeyHooks(SDL_Keycode inKeyCode, unsigned int inModifier, const std::vector<KeyHook<CONTEXT>>& inKeyHookList, ARGS... inArgs)
	{
		for (const auto& key_hook : inKeyHookList)
		{
			if (key_hook.TryConsume(inKeyCode, inModifier, inArgs...))
				return true;
		}
		return false;
	}


	template<typename CONTEXT>
	KeyHook<CONTEXT>::KeyHook()
		: m_KeyCode(0)
		, m_Modifiers(0)
	{

	}


	template<typename CONTEXT>
	KeyHook<CONTEXT>::KeyHook(const std::string& inIdentifier, const KeyHookStore& inKeyHookStore, std::function<CONTEXT> inKeyDownCallback)
		: m_Identifier(inIdentifier)
		, m_KeyDownCallback(inKeyDownCallback)
	{
		const auto& key = inKeyHookStore.GetKey(inIdentifier);

		m_KeyCode = key.m_KeyCode;
		m_Modifiers = key.m_Modifiers;
	}


	template<typename CONTEXT>
	KeyHook<CONTEXT>::KeyHook(const std::string& inIdentifier, SDL_Keycode inKeyCode, unsigned int inModifier, std::function<CONTEXT> inKeyDownCallback)
		: m_Identifier(inIdentifier)
		, m_KeyCode(inKeyCode)
		, m_Modifiers(inModifier)
		, m_KeyDownCallback(inKeyDownCallback)
	{
	}


	template<typename CONTEXT>
	const bool KeyHook<CONTEXT>::operator == (KeyHook<CONTEXT>& inOther) const
	{
		if (m_KeyCode != inOther.m_KeyCode)
			return false;

		if (m_Modifiers != 0 || inOther.m_Modifiers != 0)
		{
			if (RequireShiftDown() != inOther.RequireShiftDown())
				return false;
			if (RequireControlDown() != inOther.RequireControlDown())
				return false;
			if (RequireAltDown() != inOther.RequireAltDown())
				return false;
		}

		return true;
	}

	template<typename CONTEXT>
	const bool KeyHook<CONTEXT>::operator != (KeyHook<CONTEXT>& inOther) const
	{
		return !(*this == inOther);
	}

	template<typename CONTEXT>
	KeyHook<CONTEXT>::operator bool() const
	{
		return m_KeyCode != 0;
	}

	template<typename CONTEXT>
	const std::string& KeyHook<CONTEXT>::GetIdentifier() const
	{
		return m_Identifier;
	}

	template<typename CONTEXT>
	void KeyHook<CONTEXT>::ConfigureKey(SDL_Keycode inKeyCode, unsigned int inModifier)
	{
		m_KeyCode = inKeyCode;
		m_Modifiers = inModifier;
	}


	template<typename CONTEXT>
	bool KeyHook<CONTEXT>::RequireShiftDown() const
	{
		return (m_Modifiers & Foundation::Keyboard::Shift) != 0;
	}

	template<typename CONTEXT>
	bool KeyHook<CONTEXT>::RequireControlDown() const
	{
		return (m_Modifiers & Foundation::Keyboard::Control) != 0;
	}

	template<typename CONTEXT>
	bool KeyHook<CONTEXT>::RequireAltDown() const
	{
		return (m_Modifiers & Foundation::Keyboard::Alt) != 0;
	}


	template<typename CONTEXT>
	bool KeyHook<CONTEXT>::TryConsume(SDL_Keycode inKeyCode, unsigned int inModifier) const
	{
		if (inKeyCode == m_KeyCode)
		{
			if (Foundation::KeyboardUtils::IsModifierExclusivelyDown(inModifier, m_Modifiers))
				return OnConsume();
		}

		return false;
	}


	template<typename CONTEXT>
	template<typename ...ARGS>
	bool KeyHook<CONTEXT>::TryConsume(SDL_Keycode inKeyCode, unsigned int inModifier, ARGS... inArgs) const
	{
		if (inKeyCode == m_KeyCode)
		{
			if (Foundation::KeyboardUtils::IsModifierExclusivelyDown(inModifier, m_Modifiers))
				return OnConsume(inArgs...);
		}

		return false;
	}


	template<typename CONTEXT>
	template<typename ...ARGS>
	bool KeyHook<CONTEXT>::OnConsume(ARGS... inArgs) const
	{
		if (m_KeyDownCallback)
			return m_KeyDownCallback(inArgs...);

		return false;
	}

}
